# This script installs RightLink assuming it has been downloaded and unpacked. It is
# called at the end of rightlink.boot.ps1, rightlink.enable.ps1 and rightlink.install.ps1
# scripts, which also may set up dependencies and credentials before calling this
# script to install RightLink
#
# Parameters:
#  -Start: autostart service after installation
#  -NoNetworkingScripts: Skip installation of RightScale network configuration scripts for RCA-V
#  -Username: Have RightLink run as user. Defaults to RightLink user.
#  -Password: Password to have RightLink run as user. Defaults to random password for RightLink user

Param(
  [switch]$Start,
  [string]$Username,
  [string]$Password,
  [switch]$NoNetworkingScripts,
  [switch]$Debug
)

if ($Debug) {
  $DebugPreference = "Continue"
}

function New-SWRandomPassword {
  <#
  .Synopsis
     Generates one or more complex passwords designed to fulfill the requirements for Active Directory
  .DESCRIPTION
     Generates one or more complex passwords designed to fulfill the requirements for Active Directory
  .EXAMPLE
     New-SWRandomPassword
     C&3SX6Kn

     Will generate four passwords, each with a length of between 8 and 12 chars.
  .EXAMPLE
     New-SWRandomPassword -InputStrings abc, ABC, 123 -PasswordLength 4
     3ABa

     Generates a password with a length of 4 containing atleast one char from each InputString
  .EXAMPLE
     New-SWRandomPassword -InputStrings abc, ABC, 123 -PasswordLength 4 -FirstChar abcdefghijkmnpqrstuvwxyzABCEFGHJKLMNPQRSTUVWXYZ
     3ABa

     Generates a password with a length of 4 containing atleast one char from each InputString that will start with a letter from
     the string specified with the parameter FirstChar
  .OUTPUTS
     [String]
  .NOTES
     Written by Simon WÃ¥hlin, blog.simonw.se
     I take no responsibility for any issues caused by this script.
  .FUNCTIONALITY
     Generates random passwords
  .LINK
     http://blog.simonw.se/powershell-generating-random-password-for-active-directory/

  #>
  [CmdletBinding(DefaultParameterSetName='FixedLength',ConfirmImpact='None')]
  [OutputType([String])]
  Param
  (
    # Specifies minimum password length
    [Parameter(Mandatory=$false,
               ParameterSetName='RandomLength')]
    [ValidateScript({$_ -gt 0})]
    [Alias('Min')]
    [int]$MinPasswordLength = 8,

    # Specifies maximum password length
    [Parameter(Mandatory=$false,
               ParameterSetName='RandomLength')]
    [ValidateScript({
            if($_ -ge $MinPasswordLength){$true}
            else{Throw 'Max value cannot be lesser than min value.'}})]
    [Alias('Max')]
    [int]$MaxPasswordLength = 12,

    # Specifies a fixed password length
    [Parameter(Mandatory=$false,
               ParameterSetName='FixedLength')]
    [ValidateRange(1,2147483647)]
    [int]$PasswordLength = 8,

    # Specifies an array of strings containing charactergroups from which the password will be generated.
    # At least one char from each group (string) will be used.
    [String[]]$InputStrings = @('abcdefghijkmnpqrstuvwxyz', 'ABCEFGHJKLMNPQRSTUVWXYZ', '23456789', '!#%&'),

    # Specifies a string containing a character group from which the first character in the password will be generated.
    # Useful for systems which requires first char in password to be alphabetic.
    [String] $FirstChar
  )
  Begin {
    Function Get-Seed{
      # Generate a seed for randomization
      $RandomBytes = New-Object -TypeName 'System.Byte[]' 4
      $Random = New-Object -TypeName 'System.Security.Cryptography.RNGCryptoServiceProvider'
      $Random.GetBytes($RandomBytes)
      [BitConverter]::ToUInt32($RandomBytes, 0)
    }
  }
  Process {
    $Password = @{}
    # Create char arrays containing groups of possible chars
    [char[][]]$CharGroups = $InputStrings

    # Create char array containing all chars
    $AllChars = $CharGroups | ForEach-Object {[Char[]]$_}

    # Set password length
    if($PSCmdlet.ParameterSetName -eq 'RandomLength')
    {
      if($MinPasswordLength -eq $MaxPasswordLength) {
        # If password length is set, use set length
        $PasswordLength = $MinPasswordLength
      } else {
        # Otherwise randomize password length
        $PasswordLength = ((Get-Seed) % ($MaxPasswordLength + 1 - $MinPasswordLength)) + $MinPasswordLength
      }
    }

    # If FirstChar is defined, randomize first char in password from that string.
    if($PSBoundParameters.ContainsKey('FirstChar')){
      $Password.Add(0,$FirstChar[((Get-Seed) % $FirstChar.Length)])
    }
    # Randomize one char from each group
    Foreach($Group in $CharGroups) {
      if($Password.Count -lt $PasswordLength) {
        $Index = Get-Seed
        While ($Password.ContainsKey($Index)){
          $Index = Get-Seed
        }
        $Password.Add($Index,$Group[((Get-Seed) % $Group.Count)])
      }
    }

    # Fill out with chars from $AllChars
    for($i=$Password.Count;$i -lt $PasswordLength;$i++) {
      $Index = Get-Seed
      While ($Password.ContainsKey($Index)){
        $Index = Get-Seed
      }
      $Password.Add($Index,$AllChars[((Get-Seed) % $AllChars.Count)])
    }
    Write-Output -InputObject $(-join ($Password.GetEnumerator() | Sort-Object -Property Name | Select-Object -ExpandProperty Value))
  }
}

function SessionIsElevated
{
  [System.Security.Principal.WindowsPrincipal]$currentPrincipal = `
    New-Object System.Security.Principal.WindowsPrincipal(
      [System.Security.Principal.WindowsIdentity]::GetCurrent());

  [System.Security.Principal.WindowsBuiltInRole]$administratorsRole = `
    [System.Security.Principal.WindowsBuiltInRole]::Administrator;

  return $currentPrincipal.IsInRole($administratorsRole)
}

function RightscaleServiceExists
{
  $rightscale_exists = sc.exe query rightlink | select-string SERVICE_NAME
  if ($rightscale_exists) {
    return $True
  } else {
    return $False
  }
}

function LogWrite([string]$logString, [boolean]$noNewlineFlag = $False)
{
  if ($noNewlineFlag -eq $True) {
    [System.IO.File]::AppendAllText($INSTALL_LOG_FILE, $logString, [System.Text.Encoding]::UTF8)
    Write-Host -NoNewline $logString
  } else {
    Add-Content $INSTALL_LOG_FILE "$logString`n"
    Write-Host $logString
  }
}

function LogError([string]$logString)
{
  Add-Content $INSTALL_LOG_FILE "`n$logString`n"
  throw $logString
}

# Strings returned from a command like nssm.exe have extra whitespace between each
# character if assigned to a powershell var -- clear those out.
function FixWhitespace([string]$str)
{
  $str -replace "`0", ""
}

###############################################
# Begin Main
###############################################

$RIGHTLINK_DIR        = "$env:ProgramFiles\RightScale\RightLink"
$LOG_DIR              = "$RIGHTLINK_DIR\Logs"
$RIGHTLINK            = "$RIGHTLINK_DIR\rightlink.exe"
$NSSM                 = "$RIGHTLINK_DIR\nssm.exe"
$VMTOOLSD             = "$env:ProgramFiles\VMware\VMware Tools\vmtoolsd.exe"
$USERDATA_FETCHER     = "$RIGHTLINK_DIR\userdata-fetcher.ps1"
$RSC                  = "$RIGHTLINK_DIR\rsc.exe"
$RS_DIR               = "$env:ProgramData\RightScale\RightLink"
$RIGHTLINK_LOG_FILE   = "$LOG_DIR\rightlink.log"
$INSTALL_LOG_FILE     = "$LOG_DIR\install.log"
$RS_ID_FILE           = "$RS_DIR\rightscale-identity"
$RIGHTLINK_INST_DIR   = "$env:TEMP\RightLinkInstaller"

# Create RightLink and RightLink/Logs directory
New-Item -Path $RIGHTLINK_DIR -Type Directory -Force | Out-Null
New-Item -Path $LOG_DIR -Type Directory -Force | Out-Null

# If log files don't exist then create them
if (!(Test-Path -Path $INSTALL_LOG_FILE)) {
  New-Item -Path $INSTALL_LOG_FILE -Type File -Force | Out-Null
}
if (!(Test-Path -Path $RIGHTLINK_LOG_FILE)) {
  New-Item -Path $RIGHTLINK_LOG_FILE -Type File -Force | Out-Null
}

$currentTime = Get-Date
LogWrite "$currentTime Starting RightLink 10.6.4 install."

if (!(SessionIsElevated)) {
  LogError "You must run this from an elevated command prompt."
}

if (!(Test-Path -Path $RIGHTLINK_INST_DIR)) {
  LogError "Expected RightLink package to be expanded into directory $RIGHTLINK_INST_DIR, but it's missing."
}

LogWrite "Installing rsc to $RSC"
Copy-Item -Force "$RIGHTLINK_INST_DIR\RightLink\rsc.exe" "$RSC"
$RSC_VERSION = & $RSC --version
LogWrite "$RSC_VERSION command line tool installed."

LogWrite "Installing RightLink to $RIGHTLINK_DIR"
Copy-Item -Force "$RIGHTLINK_INST_DIR\RightLink\rightlink.exe" "$RIGHTLINK"
Copy-Item -Force "$RIGHTLINK_INST_DIR\RightLink\nssm.exe" "$NSSM"
Copy-Item -Force "$RIGHTLINK_INST_DIR\RightLink\userdata-fetcher.ps1" "$USERDATA_FETCHER"

#Add the new path to the $p variable. Begin with a semi-colon separator.
$p = $env:Path
if ($p -notmatch "RightLink") {
  LogWrite "Adding $RIGHTLINK_DIR to machines Path"
  $p += ";$RIGHTLINK_DIR"
  [Environment]::SetEnvironmentVariable("Path", $p, [EnvironmentVariableTarget]::Process)
  [Environment]::SetEnvironmentVariable("Path", $p, [EnvironmentVariableTarget]::Machine)
}

$status = FixWhiteSpace(& $NSSM status RightLink 2> $null)
if (!($status -like "SERVICE_*")) {

  $Computer     = [ADSI]"WinNT://$ENV:COMPUTERNAME,Computer"
  $ComputerName = $Computer.Name
  $Domain       = $ComputerName

  if ($Username -eq "") {
    LogWrite "Creating RightLink service account"
    $Username = "RightLink"
    $Password = New-SWRandomPassword -MinPasswordLength 8 -MaxPasswordLength 12
    $RightLinkUser = $Computer.Create("User", $Username)
    $RightLinkUser.SetPassword($Password)
    $RightLinkUser.SetInfo()
    $RightLinkUser.FullName = "RightLink Service User"
    $RightLinkUser.SetInfo()
    # https://msdn.microsoft.com/en-us/library/aa772300(v=vs.85).aspx
    # ADS_UF_PASSWD_CANT_CHANGE + ADS_UF_DONT_EXPIRE_PASSWD
    $RightLinkUser.UserFlags = 64 + 65536
    $RightLinkUser.SetInfo()

    $AdminGroup = [ADSI]"WinNT://$ComputerName/Administrators,group"
    $AdminGroup.Add($RightLinkUser.Path)
  } else {
    LogWrite "Use provided service account"
    $parts = $Username.Split('\')
    if ($parts.Length -eq 2) {
      $Domain = $parts[0]
      $Username = $parts[1]
    }
  }

  # We don't control permissions on individual files, we lock down permissions on
  # the containing directories and rely on inheritance rules. This removes the
  # need for the agent itself to need to worry about permissions, as long as it
  # works in the directories below.
  LogWrite "Retrieving user and group SIDs"
  $USER_SID = (Get-WmiObject -Class Win32_UserAccount -Filter "Name='${Username}' and Domain='${Domain}'").SID
  $ADMIN_SID = "S-1-5-32-544"

  LogWrite "Updating folder permissions"
  icacls $RS_DIR /inheritance:r /grant "*${USER_SID}:(OI)(CI)F" | Out-Null
  icacls $RS_DIR /inheritance:r /grant "*${ADMIN_SID}:(OI)(CI)F" | Out-Null
  icacls $LOG_DIR /inheritance:r /grant "*${USER_SID}:(OI)(CI)F" | Out-Null
  icacls $LOG_DIR /inheritance:r /grant "*${ADMIN_SID}:(OI)(CI)F" | Out-Null

  # SERVICE_RUNNING or SERVICE_STOPPED if installed, "" if not
  LogWrite "Creating Service RightLink"
  & $NSSM install RightLink $RIGHTLINK

  # NSSM has required functionality to assign user for the service.
  & $NSSM set RightLink ObjectName "${Domain}\${Username}" "${Password}"
  $status = FixWhiteSpace(& $NSSM status RightScaleUserData 2> $null)
  if (!($status -like "SERVICE_*")) {
    LogWrite "Creating RightScale user-data fetching service"
    & $NSSM install RightScaleUserData "$env:SystemRoot\System32\WindowsPowerShell\v1.0\powershell.exe" "-ExecutionPolicy Bypass -File `"`"`"$USERDATA_FETCHER`"`"`""
    & $NSSM set RightScaleUserData ObjectName "${Domain}\${Username}" "${Password}"
  }
}

& $NSSM set RightLink AppStdout $RIGHTLINK_LOG_FILE
& $NSSM set RightLink AppStderr $RIGHTLINK_LOG_FILE

# There are three ways this script is invoked here.
#   1) We're bootstrapping this using the Cloud's native bootstrapping service,
#      such as Ec2ConfigService, Azures Extensions, or Rackspace's userscripts.
#      For Ec2, the cloud lays down the identity. For all other clouds, that
#      data is still in the userdata and we still have to go fetch it. Hence
#      we start the RightScaleUserdata service which will start RightLink
#   2) Instance enablement case. We install and add a Powershell script rightlink.enable.ps1
#      that users can call to generate this file and lay it down.
#   3) We're doing a standalone install. No RS_ID_FILE will probably exist, should
#      be ok with that and not start the service at the end ($Start is False)

LogWrite "RightLink has been installed into $RIGHTLINK_DIR and installed as service RightLink."

# RightScaleUserData service will start RightLink service as Windows doesn't
# really have good dependency management of services.
& $NSSM set RightScaleUserData Start SERVICE_DELAYED_AUTO_START
& $NSSM set RightLink Start SERVICE_DEMAND_START

if ((Test-Path -Path $VMTOOLSD)) {
  if ($NoNetworkingScripts) {
    $rightnetworkDir = "$env:ProgramFiles\RightScale\RightNetwork"
    if (Test-Path -Path "$rightnetworkDir\uninstall.ps1") {
      LogWrite "Removing RightScaleNetwork service for RCA-V"
      powershell.exe -ExecutionPolicy unrestricted -File "$rightnetworkDir\uninstall.ps1"
    } else {
      LogWrite "Skipping install of RightScaleNetwork service for RCA-V"
    }
  } else {
    LogWrite "Installing RightScale Network Service for RCA-V"
    powershell.exe -ExecutionPolicy unrestricted -File "$RIGHTLINK_INST_DIR\RightNetwork\install.ps1"
  }
}

# Start RightLink
LogWrite "Checking for RightScaleUserData service ... " $True

$status = FixWhiteSpace(& $NSSM status RightScaleUserData 2> $null)
if ($status -like "*RUNNING*") {
  LogWrite "running"
} elseif ($status -like "SERVICE_*") {
  if ($Start) {
    LogWrite "installed and starting"
    & $NSSM start RightScaleUserData
  } else {
    LogWrite "installed"
  }
} else {
  LogError "RightLink service was not installed successfully!"
}

$currentTime = Get-Date
LogWrite "$currentTime RightLink 10.6.4 install complete."
