function Expand-Zipfile($zipFile, $targetDir, $basePath = $Null)
{
  # Check for the compression .NET Assembly in case it is needed to extract the zip file since Shell.Application may not
  # be available such as on Windows Core
  try {
    Add-Type -AssemblyName System.IO.Compression.FileSystem -ErrorAction Stop
    $dotnetCompression = $true
  } catch {
    $dotnetCompression = $false
  } finally {
    if ($dotnetCompression) {
      [System.Reflection.Assembly]::LoadWithPartialName("System.IO.Compression.FileSystem") | Out-Null
      [System.IO.Compression.ZipFile]::ExtractToDirectory($zipFile, $targetDir)
    } else {
      if (! $basePath) { $basePath =  (Get-ChildItem $zipFile | Select-Object -Property FullName).FullName }
      $shell = New-Object -com Shell.Application
      $zip = $shell.NameSpace($zipFile)

      if (!(Test-Path -Path $targetDir)) {
        New-Item -Path $targetDir -Type directory
      }
      foreach($item in $zip.items())
      {
        $newPath = $item.path -replace [RegEx]::Escape($basePath), $targetDir
        if ($item.Type -eq "File Folder") {
          New-Item -Path $newPath -Type directory
          Expand-Zipfile $item.Path $targetDir $basePath
        } else {
          $newDir = [io.path]::GetDirectoryName($newPath)
          $shell.Namespace($newDir).CopyHere($item)
        }
      }
    }
  }
}

function LogWrite([string]$logString, [boolean]$noNewlineFlag = $False)
{
  if ($noNewlineFlag -eq $True) {
    [System.IO.File]::AppendAllText($INSTALL_LOG_FILE, $logString, [System.Text.Encoding]::UTF8)
    Write-Host -NoNewline $logString
  } else {
    Add-Content $INSTALL_LOG_FILE "$logString`n"
    Write-Host $logString
  }
}

function LogError([string]$logString)
{
  Add-Content $INSTALL_LOG_FILE "`n$logString`n"
  throw $logString
}

###############################################
# Begin Main
###############################################

$RIGHTLINK_DIR         = "$env:ProgramFiles\RightScale\RightLink"
$ID_FILE               = "$env:ProgramData\RightScale\RightLink\rightscale-identity"
$LOG_DIR               = "$RIGHTLINK_DIR\Logs"
$INSTALL_LOG_FILE      = "$LOG_DIR\install.log"
$RIGHTLINK             = "$RIGHTLINK_DIR\rightlink.exe"
$RSC                   = "$RIGHTLINK_DIR\rsc.exe"
$RIGHTLINK_INST_DIR    = "$env:TEMP\RightLinkInstaller"
$RIGHTLINK_ZIP         = "rightlink.zip"
$RIGHTLINK_ZIP_URL     = "https://rightlink.rightscale.com/rll/10.6.4/$RIGHTLINK_ZIP"
$RIGHTLINK_INSTALLER   = "$RIGHTLINK_INST_DIR\RightLink\install.ps1"

# If log file doesn't exist then create it and the directory
New-Item -Path $LOG_DIR -Type Directory -Force | Out-Null
if (!(Test-Path -Path $INSTALL_LOG_FILE)) {
  New-Item -Path $INSTALL_LOG_FILE -Type file -Force | Out-Null
}

$currentTime = Get-Date
LogWrite "$currentTime Starting Launch-Time RightLink 10.6.4 install."

if ((Test-Path $RIGHTLINK) -and (Test-Path $RSC)) {
  LogWrite "RightLink is already installed in $RIGHTLINK"
  LogWrite "Skipping re-installation. RightLink Service should already be installed."
  exit 0
}

Remove-Item $RIGHTLINK_INST_DIR -ErrorAction SilentlyContinue -Force -Recurse
New-Item -Path $RIGHTLINK_INST_DIR -Force -Type directory | Out-Null
$wc = New-Object System.Net.WebClient
if (Test-Path $ID_FILE) {
  $proxy = ((Get-Content $ID_FILE) -Match "http_proxy=" | Out-String).Trim() -replace "http_proxy=",""
  if ($proxy) {
    LogWrite "Using proxy $proxy"
    $wp = New-Object System.Net.WebProxy
    $parts = $proxy -replace "http://","" -replace "https://","" -split "@"
    if ($parts.Length -gt 1) {
      $userparts = $parts[0] -split ":"
      $user = $userparts[0]
      $passwd = $userparts[1]
      $creds = New-Object System.Net.NetworkCredential($user, $passwd)
      $wp.Credentials = $creds
      $wp.Address = "http://" + $parts[1]
    } else {
      $wp.Address = "http://" + $parts[0]
    }
    $wc.Proxy = $wp
  }
}
$wc.DownloadFile($RIGHTLINK_ZIP_URL, "$RIGHTLINK_INST_DIR\$RIGHTLINK_ZIP")
Expand-ZipFile "$RIGHTLINK_INST_DIR\$RIGHTLINK_ZIP" $RIGHTLINK_INST_DIR | Out-Null
powershell.exe -ExecutionPolicy unrestricted -File $RIGHTLINK_INSTALLER -Start -NoNetworkingScripts
