# This script disables a "RightScale server" that had previously been a running instance and
# was enabled into a "RightScale server" using the rightlink.enable.ps1 script.
# It does this in 4 steps: (1) removes RightLink from the instance (2) removes the RightLink
# service and removes RightLink directory from the machines path (3) removes any files placed
# by RightLink (4) disable the server by making an disable call to the API
# This script is deliberately written in powershell because (a) it works on all Windows distros
# and (b) almost everyone can figure out how to make modifications.

Param(
  [alias('k')]
  [string]$refreshToken,
  [alias('s')]
  [string]$serverHref,
  [alias('a')]
  [string]$apiServer,
  [alias('h')]
  [switch]$help,
  [alias('f')]
  [switch]$force,
  [switch]$debug
)

function GetHelp
{
  Write-Host "This script will disable a RightScale server that had been a previously enabled server."
  Write-Host "Parameters:"
  Write-Host "  -RefreshToken         RightScale API refresh token from the dash Settings>API Credentials (required)"
  Write-Host "  -ServerHref           RightScale API server HREF (disables self-detection) (ex. /api/servers/123456789)"
  Write-Host "  -ApiServer            Hostname for the RightScale API, Default: $DEFAULT_SERVER"
  Write-Host "  -Force                Override command prompt confirmation for disablement"
  Write-Host "  -Debug                Verbose output"
  Write-Host "  -Help                 Display help"
  Write-Host ""
  Write-Host "Required Inputs: -RefreshToken"
  Write-Host ""
}

if ($Debug) {
  $DebugPreference = "Continue"
}

function SessionIsElevated
{
  [System.Security.Principal.WindowsPrincipal]$currentPrincipal = `
    New-Object System.Security.Principal.WindowsPrincipal(
      [System.Security.Principal.WindowsIdentity]::GetCurrent());

  [System.Security.Principal.WindowsBuiltInRole]$administratorsRole = `
    [System.Security.Principal.WindowsBuiltInRole]::Administrator;

  return $currentPrincipal.IsInRole($administratorsRole)
}

function LogWrite([string]$logString, [boolean]$noNewlineFlag = $False)
{
  if ($noNewlineFlag -eq $True) {
    [System.IO.File]::AppendAllText($INSTALL_LOG_FILE, $logString, [System.Text.Encoding]::UTF8)
    Write-Host -NoNewline $logString
  } else {
    Add-Content $INSTALL_LOG_FILE "$logString`n"
    Write-Host $logString
  }
}

function LogError([string]$logString)
{
  Add-Content $INSTALL_LOG_FILE "`n$logString`n"
  throw $logString
}

###############################################
# Begin Main
###############################################

$DEFAULT_SERVER     = "my.rightscale.com"
$RIGHTLINK_DIR      = "$env:ProgramFiles\RightScale\RightLink"
$LOG_DIR            = "$RIGHTLINK_DIR\Logs"
$INSTALL_LOG_FILE   = "$LOG_DIR\install.log"
$RIGHTLINK          = "$RIGHTLINK_DIR\rightlink.exe"
$RSC                = "$RIGHTLINK_DIR\rsc.exe"
$NSSM               = "$RIGHTLINK_DIR\nssm.exe"
$USERDATA_FETCHER   = "$RIGHTLINK_DIR\userdata-fetcher.ps1"
$RS_DIR             = "$env:ProgramData\RightScale\RightLink"
$RS_ID_FILE         = "$RS_DIR\rightscale-identity"
$RL_SECRET          = "$RS_DIR\secret"
$RL_STATE           = "$RS_DIR\state"
$TMP_DIR            = "$env:TEMP"
$RIGHTLINK_INST_DIR = "$TMP_DIR\RightLinkInstaller"
$RIGHTNETWORK_DIR   = "$env:ProgramFiles\RightScale\RightNetwork"

$currentTime = Get-Date
LogWrite "$currentTime Starting Disablement"

# If log file doesn't exist then create it and the directory
if (!(Test-Path -Path $INSTALL_LOG_FILE)) {
  New-Item -Path $INSTALL_LOG_FILE -Type File -Force | Out-Null
}

if ($help) {
  GetHelp
  exit 0
}

if (!(SessionIsElevated)) {
  LogError "ERROR: This must be run in an elevated command prompt."
}

if (!$refreshToken) {
  GetHelp
  LogError "ERROR: -RefreshToken missing."
}

if (!$apiServer) {
  if ($RS_ID_FILE) {
    $results = Select-String -Path $RS_ID_FILE -Pattern 'api_hostname=(.*)'
    $apiServer = $results.Matches[0].Groups[1].Value
  } elseif ($DEFAULT_SERVER) {
    $apiServer = $DEFAULT_SERVER
    LogWrite "Using API Host $apiServer"
  } else {
    GetHelp
    LogError "ERROR: -ApiServer missing."
  }
}

# Check for Server HREF using user-data to make an api call
if (!$serverHref) {
  if ($RS_ID_FILE) {
    if (Get-Process | Where-Object { $_.Name -eq "rightlink" }) {
      LogWrite "Locating Server HREF using user-data on file ... " $True
      $serverHref = (& $RSC --rl10 --x1 ':has(.rel:val(\"parent\")).href' cm15 `
                    index_instance_session "/api/sessions/instance" 2> $null)
      if ($serverHref) {
        LogWrite "serverHref = $serverHref"
      } else {
        LogWrite "serverHref not found."
      }
    }
  }
}

# Check for Server HREF in the log file created by enable script
if (!$serverHref) {
  if ($INSTALL_LOG_FILE) {
    LogWrite "Locating Server HREF from log files ... " $True
    $results = Select-String -Path $INSTALL_LOG_FILE -Pattern '(Created server: |Checking to see if instance is already enabled \.\.\. )serverHref = (/api/servers/[0-9]+)'
    $serverHref = $results[-1].Matches.Groups[2].Value
    if ($serverHref) {
      LogWrite "serverHref = $serverHref"
    } else {
      LogWrite "serverHref not found."
    }
  }
}

# Take out the /deployments/123456789 from the HREF if it's there
$serverHref = $serverHref -replace '/deployments/[0-9]+', ''

# Check to make sure serverHref exists and that it is valid
if (!$serverHref) {
  GetHelp
  LogError "ERROR: -ServerHref missing."
} elseif ($serverHref -notmatch '/api/servers/[0-9]+') {
  GetHelp
  LogError "ERROR: -ServerHref malformed."
}

# Querying details about current server
LogWrite "Finding current server ... " $True
$json = & $RSC --key $refreshToken --host $apiServer cm15 show $serverHref
if (!$json) {
  LogError "ERROR: Server not found."
}

# Make sure that the server is in a valid state to be disabled
$state = (Write-Output $json | & $RSC --x1 '.state' json 2> $null)
if ($state -ne "operational" -and $state -ne "booting" -and $state -notmatch 'stranded') {
  LogError "ERROR: Server found, but in the wrong state: $state."
} else {
  LogWrite "Server found and is in state: $state."
  # Store server name
  $serverName = (Write-Output $json | & $RSC --x1 '.name' json 2> $null)
}

LogWrite ""
LogWrite "The following will be deleted locally:"

# Locate RightLink
if (Test-Path -Path $RIGHTLINK) {
  LogWrite "RightLink           $RIGHTLINK"
}
# Locate RSC
if (Test-Path -Path $RSC) {
  LogWrite "RSC                 $RSC"
}
if (Test-Path -Path $NSSM) {
  LogWrite "NSSM                $NSSM"
}
if (Test-Path -Path $USERDATA_FETCHER) {
  LogWrite "User-Data Fetcher   $USERDATA_FETCHER"
}
if (Test-Path -Path $RS_ID_FILE) {
  LogWrite "User-Data           $RS_ID_FILE"
}
if (Test-Path -Path $RL_SECRET) {
  LogWrite "Secret              $RL_SECRET"
}
if (Test-Path -Path $RL_STATE) {
  LogWrite "State               $RL_STATE"
}
# Locate temporary files
if (Test-Path -Path $TMP_DIR\attachments) {
  LogWrite "Attachments         $TMP_DIR\rightlink\attachments"
}
if (Test-Path -Path $TMP_DIR\cookbooks) {
  LogWrite "Cookbooks           $TMP_DIR\rightlink\cookbooks"
}
if (Test-Path -Path $TMP_DIR\scripts) {
  LogWrite "Scripts             $TMP_DIR\rightlink\scripts"
}
if (Test-Path -Path $RIGHTLINK_INST_DIR) {
  LogWrite "RL10 Installer      $RIGHTLINK_INST_DIR"
}
if (Test-Path -Path $RIGHTNETWORK_DIR\uninstall.ps1) {
  LogWrite "Networking Scripts  $RIGHTNETWORK_DIR"
}
if (Get-WmiObject -Class Win32_UserAccount -Filter "Name='RightLink' and Domain='${ENV:COMPUTERNAME}'") {
  LogWrite "User RightLink"
}

# Prompt user for confirmation
LogWrite ""
LogWrite "The following Server will be RightScale disabled:"
LogWrite "Server Name         $serverName"
LogWrite "Server HREF         $serverHref"

if ($force) {
  $response = "y"
} else {
  LogWrite ""
  $response = Read-Host "Do you want to proceed? [y/N]"
}

LogWrite ""

if ($response -match '^([yY][eE][sS]|[yY])$') {
  # Get cloud href from current instance href
  $currentInstanceHref = ($json | & $RSC --x1 ':has(.rel:val(\"current_instance\")).href' json)
  $cloudHref = ($currentInstanceHref -replace '/instances/.+$', '')
  # Get cloud type
  $cloudType = (& $RSC --key $refreshToken --host $apiServer --x1 '.cloud_type' cm15 show $cloudHref)
  if ($cloudType -ne 'uca') {
    # Execute disable call
    $instanceHref = (& $RSC --key $refreshToken --host $apiServer --xh 'location' cm15 unwrap $serverHref 2>&1)
  }
  # Confirm disable success / failure
  if (($cloudType -ne 'uca') -and ($instanceHref -notmatch '^/api/clouds/[0-9]+/instances/.*$')) {
    # If server failed to disable, output response from API
    LogWrite "ERROR: Failed to disable server: $serverHref"
    LogError "$instanceHref"
  } else {
    # Remove RightScale environment variables
    $p = $env:Path -replace [Regex]::Escape(";$env:ProgramFiles\RightScale\RightLink"), ''
    [Environment]::SetEnvironmentVariable("Path", $p, [EnvironmentVariableTarget]::Machine)
    # Stop NSSM so it doesn't start RightLink again
    Stop-Process -ErrorAction SilentlyContinue -Force -Name 'nssm'
    # Stop RightLink without running decommission bundle
    Stop-Process -ErrorAction SilentlyContinue -Force -Name 'rightlink'
    if ($cloudType -eq 'uca') {
      & $RSC --key $refreshToken --host $apiServer cm15 terminate $serverHref
      while ((& $RSC --key $refreshToken --host $apiServer --x1 '.state' cm15 show $serverHref) -ne 'inactive') {
        Start-Sleep 1
      }
      & $RSC --key $refreshToken --host $apiServer cm15 destroy $serverHref
    }
    # Remove RightLink services
    & $NSSM remove RightLink confirm
    & $NSSM remove RightScaleUserData confirm
    # Remove rightscale_network components
    if (Test-Path -Path $RIGHTNETWORK_DIR\uninstall.ps1) {
      LogWrite "Removing RightScale networking scripts"
      powershell.exe -ExecutionPolicy unrestricted -File "$RIGHTNETWORK_DIR\uninstall.ps1"
    }
    # Delete RightLink
    Remove-Item -ErrorAction SilentlyContinue -Force $RIGHTLINK
    # Delete NSSM
    Remove-Item -ErrorAction SilentlyContinue -Force $NSSM
    # Delete RSC
    Remove-Item -ErrorAction SilentlyContinue -Force $RSC
    # Delete RightScale user-data, user-data fetcher, Secret and State files
    Remove-Item -ErrorAction SilentlyContinue -Force $RS_ID_FILE
    Remove-Item -ErrorAction SilentlyContinue -Force $RL_SECRET
    Remove-Item -ErrorAction SilentlyContinue -Force $RL_STATE
    Remove-Item -ErrorAction SilentlyContinue -Force $USERDATA_FETCHER
    # Delete Attachments, Cookbooks, Scripts, and Temporary files
    Remove-Item -Recurse -ErrorAction SilentlyContinue -Force $TMP_DIR\rightlink
    Remove-Item -Recurse -ErrorAction SilentlyContinue -Force $RIGHTLINK_INST_DIR
    # Delete RightLink User
    if (Get-WmiObject -Class Win32_UserAccount -Filter "Name='RightLink'") {
      $ADSIComp = [ADSI]"WinNT://$env:COMPUTERNAME"
      $ADSIComp.Delete('User','RightLink')
    }
    if ($cloudType -ne 'uca') {
      LogWrite ""
      LogWrite "The disabled instance is located at:"
      LogWrite "Instance HREF       $instanceHref"
    }
  }
} else {
  LogWrite "Exiting script with no changes."
  exit 0
}

LogWrite ""
$currentTime = Get-Date
LogWrite "$currentTime Disablement complete.  Server has been reverted to pre-enablement state."
