# This script will install/upgrade RightLink. This script is meant to be used to create
# custom images with RightLink pre-installed, such as when creating private cloud
# images or snapshotting existing images. If RightLink already exists on the instance,
# this script will replace it with version 10.6.4.
#
# This script depends on .NET framework (3.5+). PowerShell 2.0+ is also required.
#
# For detailed instruction, see http://docs.rightscale.com/rl10/reference/rl10_install.html

# If a Username/Password to a local or domain user is supplied, this install script
# will attempt to run RightLink as that user. If no Username/Password is supplied,
# a local Administrator called "RightLink" with a randomly generated password will be created
# for RightLink to run under.

Param(
  [alias('h')][switch]$Help,
  [alias('v')][switch]$Debug,
  [alias('u')][string]$Username,
  [alias('p')][string]$Password,
  [alias('x')][string]$Proxy,
  [alias('n')][switch]$NoNetworkingScripts
)

function GetHelp
{
  Write-Host "Installs/Upgrades RightLink on an an image."
  Write-Host "Parameters:"
  Write-Host "  -Username             RightLink Service User Name (default: RightLink)"
  Write-Host "  -Password             RightLink Service User Password (default: Randomly generated password)"
  Write-Host "  -NoNetworkingScripts  Skip installation of RightScale network configuration scripts for RCA-V"
  Write-Host "  -Proxy                Install Rightlink through HTTP proxy"
  Write-Host "  -Debug                Verbose output"
  Write-Host "  -Help                 Display help"
  Write-Host ""
}

function Test-Credential([string]$domain, [string]$username, [string]$password) {
  $LogonUserDefinition = @'
  [DllImport("advapi32.dll", SetLastError=true)]
  public static extern bool LogonUser(
      string lpszUsername,
      string lpszDomain,
      string lpszPassword,
      int dwLogonType,
      int dwLogonProvider,
      out IntPtr phToken
  );
'@
  $Advapi32 = Add-Type -MemberDefinition $LogonUserDefinition -Name 'Advapi32' -Namespace 'Win32' -PassThru
  $LOGON32_LOGON_NETWORK = 3
  $LOGON32_PROVIDER_DEFAULT = 0

  $token = New-Object -TypeName IntPtr
  $Advapi32::LogonUser($username, $domain, $password, $LOGON32_LOGON_NETWORK, $LOGON32_PROVIDER_DEFAULT, [ref]$token)
}

function Expand-Zipfile($zipFile, $targetDir, $basePath = $Null)
{
  # Check for the compression .NET Assembly in case it is needed to extract the zip file since Shell.Application may not
  # be available such as on Windows Core
  try {
    Add-Type -AssemblyName System.IO.Compression.FileSystem -ErrorAction Stop
    $dotnetCompression = $true
  } catch {
    $dotnetCompression = $false
  } finally {
    if ($dotnetCompression) {
      [System.Reflection.Assembly]::LoadWithPartialName("System.IO.Compression.FileSystem") | Out-Null
      [System.IO.Compression.ZipFile]::ExtractToDirectory($zipFile, $targetDir)
    } else {
      if (! $basePath) { $basePath =  (Get-ChildItem $zipFile | Select-Object -Property FullName).FullName }
      $shell = New-Object -com Shell.Application
      $zip = $shell.NameSpace($zipFile)

      if (!(Test-Path -Path $targetDir)) {
        New-Item -Path $targetDir -Type directory
      }
      foreach($item in $zip.items())
      {
        $newPath = $item.path -replace [RegEx]::Escape($basePath), $targetDir
        if ($item.Type -eq "File Folder") {
          New-Item -Path $newPath -Type directory
          Expand-Zipfile $item.Path $targetDir $basePath
        } else {
          $newDir = [io.path]::GetDirectoryName($newPath)
          $shell.Namespace($newDir).CopyHere($item)
        }
      }
    }
  }
}

if ($Help) {
  GetHelp
  exit 0
}

if ($Debug) {
  $DebugPreference = "Continue"
}

###############################################
# Begin Main
###############################################

$RIGHTLINK_DIR       = "$env:ProgramFiles\RightScale\RightLink"
$LOG_DIR             = "$RIGHTLINK_DIR\Logs"
$INSTALL_LOG_FILE    = "$LOG_DIR\install.log"
$RIGHTLINK           = "$RIGHTLINK_DIR\rightlink.exe"
$RS_DIR              = "$env:ProgramData\RightScale\RightLink"
$RIGHTLINK_LOG_FILE  = "$RS_DIR\rightlink.log"
$RS_ID_FILE          = "$RS_DIR\rightscale-identity"
$RIGHTLINK_ZIP_URL   = "https://rightlink.rightscale.com/rll/10.6.4/rightlink.zip"
$RIGHTLINK_INST_DIR  = "$env:TEMP\RightLinkInstaller"
$RIGHTLINK_INSTALLER = "$RIGHTLINK_INST_DIR\RightLink\install.ps1"

if ($Username -ne "") {
  $Domain = ""
  $CheckUsername = $Username
  $parts = $Username.Split('\', 2)
  if ($parts.Length -eq 2) {
    $Domain = $parts[0]
    $CheckUsername = $parts[1]
  }
  if ( (Test-Credential $Domain $CheckUsername $Password) -eq $false ) {
    Write-Host "Please provide a valid Username and Password"
    exit 1
  }
}

New-Item -Path $LOG_DIR -Type Directory -Force | Out-Null

# If log files don't exist then create them and the directory
if (!(Test-Path -Path $INSTALL_LOG_FILE)) {
  New-Item -Path $INSTALL_LOG_FILE -Type File -Force | Out-Null
}
if (!(Test-Path -Path $RIGHTLINK_LOG_FILE)) {
  New-Item -Path $RIGHTLINK_LOG_FILE -Type File -Force | Out-Null
}

$logString = 'Starting RightLink 10.6.4 install.'
Add-Content $INSTALL_LOG_FILE "$logString`n"
Write-Host $logString

# ===== Download and expand RightLink

if (Test-Path -Path $RIGHTLINK_INST_DIR) {
  Remove-Item $RIGHTLINK_INST_DIR -Force -Recurse
}
New-Item -Path $RIGHTLINK_INST_DIR -Force -Type Directory | Out-Null
$wc = New-Object System.Net.WebClient
$proxyAddr = (Get-ItemProperty 'HKCU:\Software\Microsoft\Windows\CurrentVersion\Internet Settings').ProxyServer
if ($proxy) {
  Write-Host "Using proxy $proxy"
  $webproxy = New-Object System.Net.WebProxy
  $parts = $proxy -replace "http://","" -replace "https://","" -split "@"
  if ($parts.Length -gt 1) {
    $userparts = $parts[0] -split ":"
    $user = $userparts[0]
    $passwd = $userparts[1]
    $creds = New-Object System.Net.NetworkCredential($user, $passwd)
    $webproxy.Credentials = $creds
    $webProxy.Address = "http://" + $parts[1]
  } else {
    $webProxy.Address = "http://" + $parts[0]
  }
  $wc.Proxy = $webproxy
} elseif ($proxyAddr) {
  # should be cautious about proxies authenticated using BASIC authentication in windows
  # https://support.microsoft.com/en-us/kb/2778122
  # http://blogs.msdn.com/b/ieinternals/archive/2012/08/03/manual-proxy-authentication-requiring-basic-or-digest-breaks-many-applications.aspx
  # So we basically only support unauthenticated proxies in this case
  Write-Host "Using system proxy $proxyAddr"
  $webproxy = [System.Net.WebRequest]::GetSystemWebProxy()
  $webproxy.Credentials = [System.Net.CredentialCache]::DefaultNetworkCredentials
  $wc.Proxy = $webproxy
  $wc.UseDefaultCredentials = $true
}
$wc.DownloadFile($RIGHTLINK_ZIP_URL, "$RIGHTLINK_INST_DIR\rightlink.zip")
Expand-Zipfile "$RIGHTLINK_INST_DIR\rightlink.zip" "$RIGHTLINK_INST_DIR" | Out-Null

if ($NoNetworkingScripts) {
  $NoNetworkingScriptsArg = "-NoNetworkingScripts"
} else {
  $NoNetworkingScriptsArg = ""
}

if ($Username -eq "") {
  powershell.exe -ExecutionPolicy unrestricted -File $RIGHTLINK_INSTALLER $NoNetworkingScriptsArg
} else {
  powershell.exe -ExecutionPolicy unrestricted -File $RIGHTLINK_INSTALLER -Username ${Username} -Password ${Password} $NoNetworkingScriptsArg
}
