# This script uninstalls a previous RightLink from disk. If the server has been
# "enabled" into a RightScale server, it won't disable it and turn it back into
# an instance. For that, see rightlink.disable.ps1.
#
# This script depends on .NET framework (3.5+). PowerShell 2.0+ is also required.

Param(
  [alias('h')]
  [switch]$help,
  [alias('f')]
  [switch]$force,
  [switch]$debug
)

function GetHelp
{
  Write-Host "This script will uninstall RightLink."
  Write-Host "Parameters:"
  Write-Host "  -Force                Override command prompt confirmation for uninstall"
  Write-Host "  -Debug                Verbose output"
  Write-Host "  -Help                 Display help"
}

if ($Debug) {
  $DebugPreference = "Continue"
}

function SessionIsElevated
{
  [System.Security.Principal.WindowsPrincipal]$currentPrincipal = `
    New-Object System.Security.Principal.WindowsPrincipal(
      [System.Security.Principal.WindowsIdentity]::GetCurrent());

  [System.Security.Principal.WindowsBuiltInRole]$administratorsRole = `
    [System.Security.Principal.WindowsBuiltInRole]::Administrator;

  return $currentPrincipal.IsInRole($administratorsRole)
}

function LogWrite([string]$logString, [boolean]$noNewlineFlag = $False)
{
  if ($noNewlineFlag -eq $True) {
    [System.IO.File]::AppendAllText($INSTALL_LOG_FILE, $logString, [System.Text.Encoding]::UTF8)
    Write-Host -NoNewline $logString
  } else {
    Add-Content $INSTALL_LOG_FILE "$logString`n"
    Write-Host $logString
  }
}

function LogError([string]$logString)
{
  Add-Content $INSTALL_LOG_FILE "`n$logString`n"
  throw $logString
}

###############################################
# Begin Main
###############################################

$RIGHTLINK_DIR      = "$env:ProgramFiles\RightScale\RightLink"
$LOG_DIR            = "$RIGHTLINK_DIR\Logs"
$INSTALL_LOG_FILE   = "$LOG_DIR\install.log"
$RIGHTLINK          = "$RIGHTLINK_DIR\rightlink.exe"
$RSC                = "$RIGHTLINK_DIR\rsc.exe"
$NSSM               = "$RIGHTLINK_DIR\nssm.exe"
$USERDATA_FETCHER   = "$RIGHTLINK_DIR\userdata-fetcher.ps1"
$RS_DIR             = "$env:ProgramData\RightScale\RightLink"
$RS_ID_FILE         = "$RS_DIR\rightscale-identity"
$RL_SECRET          = "$RS_DIR\secret"
$RL_STATE           = "$RS_DIR\state"
$TMP_DIR            = "$env:TEMP"
$RIGHTLINK_INST_DIR = "$TMP_DIR\RightLinkInstaller"
$RIGHTNETWORK_DIR   = "$env:ProgramFiles\RightScale\RightNetwork"

$currentTime = Get-Date
LogWrite "$currentTime Starting Uninstall"

# If log file doesn't exist then create it and the directory
if (!(Test-Path -Path $INSTALL_LOG_FILE)) {
  New-Item -Path $INSTALL_LOG_FILE -Type File -Force | Out-Null
}

if ($help) {
  GetHelp
  exit 0
}

if (!(SessionIsElevated)) {
  LogError "ERROR: This must be run in an elevated command prompt."
}

LogWrite ""
LogWrite "The following will be deleted locally:"

# Locate RightLink
if (Test-Path -Path $RIGHTLINK) {
  LogWrite "RightLink           $RIGHTLINK"
}
# Locate RSC
if (Test-Path -Path $RSC) {
  LogWrite "RSC                 $RSC"
}
if (Test-Path -Path $NSSM) {
  LogWrite "NSSM                $NSSM"
}
if (Test-Path -Path $USERDATA_FETCHER) {
  LogWrite "User-Data Fetcher   $USERDATA_FETCHER"
}
if (Test-Path -Path $RS_ID_FILE) {
  LogWrite "User-Data           $RS_ID_FILE"
}
if (Test-Path -Path $RL_SECRET) {
  LogWrite "Secret              $RL_SECRET"
}
if (Test-Path -Path $RL_STATE) {
  LogWrite "State               $RL_STATE"
}
# Locate temporary files
if (Test-Path -Path $TMP_DIR\attachments) {
  LogWrite "Attachments         $TMP_DIR\rightlink\attachments"
}
if (Test-Path -Path $TMP_DIR\cookbooks) {
  LogWrite "Cookbooks           $TMP_DIR\rightlink\cookbooks"
}
if (Test-Path -Path $TMP_DIR\scripts) {
  LogWrite "Scripts             $TMP_DIR\rightlink\scripts"
}
if (Test-Path -Path $RIGHTLINK_INST_DIR) {
  LogWrite "RL10 Installer      $RIGHTLINK_INST_DIR"
}
if (Test-Path -Path $RIGHTNETWORK_DIR\uninstall.ps1) {
  LogWrite "Networking Scripts  $RIGHTNETWORK_DIR"
}
if (Get-WmiObject -Class Win32_UserAccount -Filter "Name='RightLink'") {
  LogWrite "User RightLink"
}

if ($force) {
  $response = "y"
} else {
  LogWrite ""
  $response = Read-Host "Do you want to proceed? [y/N]"
}

LogWrite ""

if ($response -match '^([yY][eE][sS]|[yY])$') {
  # Remove RightScale environment variables
  $p = $env:Path -replace [Regex]::Escape(";$env:ProgramFiles\RightScale\RightLink"), ''
  [Environment]::SetEnvironmentVariable("Path", $p, [EnvironmentVariableTarget]::Machine)
  # Stop NSSM so it doesn't start RightLink again
  Stop-Process -ErrorAction SilentlyContinue -Force -Name 'nssm'
  # Stop RightLink without running decommission bundle
  Stop-Process -ErrorAction SilentlyContinue -Force -Name 'rightlink'
  # Remove RightLink services
  & $NSSM remove RightLink confirm
  & $NSSM remove RightScaleUserData confirm
  # Remove rightscale_network components
  if (Test-Path -Path $RIGHTNETWORK_DIR\uninstall.ps1) {
    LogWrite "Removing RightScale networking scripts"
    powershell.exe -ExecutionPolicy unrestricted -File "$RIGHTNETWORK_DIR\uninstall.ps1"
  }
  # Delete RightLink
  Remove-Item -ErrorAction SilentlyContinue -Force $RIGHTLINK
  # Delete NSSM
  Remove-Item -ErrorAction SilentlyContinue -Force $NSSM
  # Delete RSC
  Remove-Item -ErrorAction SilentlyContinue -Force $RSC
  # Delete RightScale user-data, user-data fetcher, Secret and State files
  Remove-Item -ErrorAction SilentlyContinue -Force $RS_ID_FILE
  Remove-Item -ErrorAction SilentlyContinue -Force $RL_SECRET
  Remove-Item -ErrorAction SilentlyContinue -Force $RL_STATE
  Remove-Item -ErrorAction SilentlyContinue -Force $USERDATA_FETCHER
  # Delete Attachments, Cookbooks, Scripts, and Temporary files
  Remove-Item -Recurse -ErrorAction SilentlyContinue -Force $TMP_DIR\rightlink
  Remove-Item -Recurse -ErrorAction SilentlyContinue -Force $RIGHTLINK_INST_DIR
  # Delete RightLink User
  if (Get-WmiObject -Class Win32_UserAccount -Filter "Name='RightLink'") {
    $ADSIComp = [ADSI]"WinNT://$env:COMPUTERNAME"
    $ADSIComp.Delete('User','RightLink')
  }
} else {
  LogWrite "Exiting script with no changes."
  exit 0
}

LogWrite ""
$currentTime = Get-Date
LogWrite "$currentTime Uninstall complete."
