# This script is installed alongside RightLink and installed as a Windows service
# called RightScaleUserdata and is dependent on .NET framework for HTTP calls.
#
# It fetches RightScale authentication tokens for use with the RightLink client through
# each cloud's userdata service, acting as a sort of cloud-init equivalent for windows.
# Each cloud's metadata/userdata service is polled simultaneously and then are checked
# in a fixed order. This script will try to fetch userdata for up to 3 minutes from
# the various services.
#
# This script is needed to support booting of RightScale Servers for various clouds
# such as RCA-V, OpenStack, CloudStack, and Softlayer.
#

Param(
  [switch]$Help,
  [switch]$Debug
)

function GetHelp
{
  Write-Host "This script will fetch cloud userdata and write out a rightscale identity file for RightLink"
  Write-Host "Parameters:"
  Write-Host "  -Debug                Verbose output"
  Write-Host "  -Help                 Display help"
  Write-Host ""
}

if ($Debug) {
  $DebugPreference = "Continue"
}

if ($Help) {
  GetHelp
  exit 0
}

function LogWrite([string]$logString) {
  $currentTime = Get-Date -Format "yyyy-MM-dd hh:mm:ss"
  Add-Content $LOG_FILE "[$currentTime] INFO $logString`n"
  Write-Host $logString
}

function LogError([string]$logString) {
  $currentTime = Get-Date -Format "yyyy-MM-dd hh:mm:ss"
  Add-Content $LOG_FILE "[$currentTime] ERR  $logString`n"
  throw $logString
}

###############################################
# Begin Main
###############################################

$RIGHTLINK_DIR    = "$env:ProgramFiles\RightScale\RightLink"
$LOG_DIR          = "$RIGHTLINK_DIR\Logs"
$LOG_FILE         = "$LOG_DIR\userdata-fetcher.log"
$NSSM             = "$RIGHTLINK_DIR\nssm.exe"
$RL               = "$RIGHTLINK_DIR\rightlink.exe"
$RS_DATADIR       = "$env:ProgramData\RightScale\RightLink"
$RS_ID_FILE       = "$RS_DATADIR\rightscale-identity"


# If RightLink directory doesn't exist then create it
New-Item -Path $RIGHTLINK_DIR -Type Directory -Force | Out-Null
New-Item -Path $LOG_DIR -Type Directory -Force | Out-Null

# If log file doesn't exist then create it and the directory
if (!(Test-Path -Path $LOG_FILE)) {
  New-Item -Path $LOG_FILE -Type File -Force | Out-Null
}

LogWrite "Fetching Cloud Userdata"

$wc = New-Object System.Net.WebClient
$wc.Proxy = [System.Net.GlobalProxySelection]::GetEmptyWebProxy()

# ===== Retrieve cloud userdata

$Ec2UserData = {
  $wc = New-Object System.Net.WebClient
  $wc.Proxy = [System.Net.GlobalProxySelection]::GetEmptyWebProxy()
  $userdata = $wc.DownloadString("http://169.254.169.254/latest/user-data")
  if ($userdata -eq $null -or $userdata -eq "") {
    return $false
  } else {
    return $userdata
  }
}

$AzureUserData = {
  # Azure service will write this file out if userdata is specified
  if (Test-Path -Path "C:\AzureData\CustomData.bin") {
    $userdata = Get-Content "C:\AzureData\CustomData.bin" | Out-String
    return $userdata
  } else {
    return $false
  }
}

$ConfigDriveUserData = {
  $candidates = [System.Io.DriveInfo]::GetDrives() | where-object {
    $_.VolumeLabel -eq "METADATA" -or
    $_.VolumeLabel -eq "config-2" -or
    $_.DriveType -eq "CDRom"
  }
  ForEach ($c in $candidates) {
    $path = "$($c.RootDirectory)openstack\latest\user_data"
    if (Test-Path -Path $path) {
      $userdata = Get-Content $path | Out-String
      return $userdata
    }
  }
  return $false
}

$CloudStackUserData = {
  $dhcp = ipconfig /all | find /i "DHCP Server"
  $dhcp -match ": (.*)" | Out-Null
  $ip = $matches[1]
  $wc = New-Object System.Net.WebClient
  $wc.Proxy = [System.Net.GlobalProxySelection]::GetEmptyWebProxy()
  $userdata = $wc.DownloadString("http://$ip/latest/user-data")
  if ($userdata -eq $null -or $userdata -eq "") {
    return $false
  } else {
    return $userdata
  }
}

$GoogleUserData = {
  $wc = New-Object System.Net.WebClient
  $wc.Proxy = [System.Net.GlobalProxySelection]::GetEmptyWebProxy()
  $wc.Headers.add('Metadata-Flavor','Google')
  $userdata = $wc.DownloadString("http://169.254.169.254/computeMetadata/v1/instance/attributes/user-data")
  if ($userdata -eq $null -or $userdata -eq "") {
    return $false
  } else {
    return $userdata
  }
}

$OpenstackUserdata = {
  $wc = New-Object System.Net.WebClient
  $wc.Proxy = [System.Net.GlobalProxySelection]::GetEmptyWebProxy()
  $userdata = $wc.DownloadString("http://169.254.169.254/openstack/latest/user_data")
  if ($userdata -eq $null -or $userdata -eq "") {
    return $false
  } else {
    return $userdata
  }
}

$SoftlayerUserData = {
  $wc = New-Object System.Net.WebClient
  $wc.Proxy = [System.Net.GlobalProxySelection]::GetEmptyWebProxy()
  $userdata = $wc.DownloadString("https://api.service.softlayer.com/rest/v3.1/SoftLayer_Resource_Metadata/getUserMetadata.txt")
  if ($userdata -eq $null -or $userdata -eq "") {
    return $false
  } else {
    return $userdata
  }
}

$ExtraConfigUserData = {
  $rpctool = "$env:ProgramFiles\VMware\VMware Tools\rpctool.exe"
  if (Test-Path -Path $rpctool) {
    $userdata = (& $rpctool "info-get guestinfo.userdata" 2> $null)
  }
  if ($userdata -eq $null -or $userdata.Length -eq 0) {
    return $false
  } else {
    return $userdata
  }
}

$DataSources = @(
  @("Softlayer", $SoftlayerUserData),
  @("ConfigDrive", $ConfigDriveUserData),
  @("ExtraConfig", $ExtraConfigUserData),
  @("Openstack", $OpenstackUserData),
  @("Azure", $AzureUserData),
  @("Ec2", $Ec2UserData),
  @("Cloudstack", $CloudstackUserData),
  @("Google", $GoogleUserData)
)


$DataSources.GetEnumerator() | Foreach-Object {
  Start-Job -Name $_[0] -ScriptBlock $_[1] | Out-Null
}

# We run all datasources simultaneously up to 10 times each, max 60 seconds per try.
# All jobs are run in parallel but we check results in the fixed order above and use
# the first successful data source that completed successfully.
$userdata = $false
$selfcheck_success = $false
For($i = 0; $i -lt 10; $i += 1) {
  # Outside programs (cloud agents) may write this file. Also we have have a valid
  # file from a previous boot. If so, cutout of this loop and move on to booting the agent
  if (Test-Path -Path $RS_ID_FILE) {
    Get-Content $RS_ID_FILE | Select-String '^$' -notmatch | ForEach-Object {
      $line = $_ -replace '\$env:', '' -replace '^RS_', ''
      if ($line -match "=") {
        $param, $value = $line -split "=", 2
        [Environment]::SetEnvironmentVariable($param, $value, 'process')
      }
    }

    & $RL -selfcheck
    if ($?) {
      LogWrite "Skipping fetching of userdata, valid credentials found on disk"
      $selfcheck_success = $true
      break
    }
  }

  Start-Sleep -Seconds 60

  $DataSources.GetEnumerator() | Foreach-Object {
    $SourceName = $_[0]
    $Job = @(Get-Job -Name $SourceName)[-1]
    # Job states can be "Completed", "Running", or "Failed". We retry on
    # Running or Failed as that can indicate temporary network errors
    if ($Job.State -eq "Completed") {
      $userdata = Receive-Job -Job $Job
      if ($userdata -ne $false) {
        LogWrite "Found $SourceName userdata $userdata"
        break # This should break out of the top level for loop
      } else {
        # Do nothing, we'll clean up all jobs and start again
        LogWrite "Could not query userdata from $SourceName"
      }
    } else {
      # valid $Job.State values are NotStarted, Running, Completed, Failed, Stopped,
      # Blocked, Suspended, Disconnected, Suspending, Stopping. Only care about
      # Completed, everything else considered a failure
      LogWrite "Could not query userdata from $SourceName"
    }
  }
  if ($userdata -ne $false) {
    break
  }
  LogWrite "All datasources failed, trying all again..."
  $DataSources.GetEnumerator() | Foreach-Object {
    Start-Job -Name $_[0] -ScriptBlock $_[1] | Out-Null
  }
}

if ($userdata -eq $false) {
  # In the case on enablement or RightLink being installed on the fly via a
  # external agent such as Ec2ConfigService, we don't care if we didn't find
  # any appropriate userdata, we should have good userdata already, so we continue
  # on our way
  if ($selfcheck_success -eq $false) {
    LogWrite "WARNING: No valid userdata sources found"
  }
} else {
  # ===== Save user-data in a file
  # We save the user-data in a file so it remains available if RightLink10 is restarted,
  # for example due to a reboot. Here we handle query-string user-data format, which we may have
  # to switch to mime depending on the MCIs we want to use and to support relaunching the
  # instance.

  if (!(Test-Path -Path $RS_DATADIR)) {
    New-Item -Path $RS_DATADIR -Type Directory -Force | Out-Null
  }

  # Theres three different formats:
  #   1. RightLink: querystring format
  #   2. RightLinkLite: mime format
  #   3. Cloudbase-init format: a powershell script starting with #ps1_sysnative
  # For options 1 and 2 we want to extract the important rightscale auth tokens and
  # write them to disk. For option 3 we just want to execute whatever is there --
  # unless cloudbase-init is installed, in which case we defer to that.
  $userDataLines = $userData | foreach {$_ -split "&|`n"}

  $RS_VARS = ""
  foreach ($line in $userDataLines) {
    $line = $line -replace 'RS_rn_auth',     'client_secret'
    $line = $line -replace 'RS_rn_id',       'client_id'
    $line = $line -replace 'RS_server',      'api_hostname'
    $line = $line -replace 'http_no_proxy',  'no_proxy'
    $line = $line -replace "(.*)='(.*)'", "`$1=`$2"
    $line = $line -replace "^RS_(.*)=(.*)", "`$1=`$2"
    if ($line -match "^(expected_public_ip|client_secret|client_id|api_hostname|account|http_proxy|no_proxy)=") {
      $RS_VARS = $RS_VARS + $line + "`n"
    }
  }

  if ($RS_VARS -ne "") {
    LogWrite "Writing RightLink enrollment info to $RS_ID_FILE"
    Set-Content $RS_ID_FILE $RS_VARS -Force
  } else {
    LogWrite "WARNING: no valid RightScale userdata was found."
  }
}

# Start RightLink service
if (Test-Path -Path $RS_ID_FILE) {
  LogWrite "Starting RightLink service"
  $RS_VARS = Get-Content $RS_ID_FILE | Select-String '^$' -notmatch | ForEach-Object {
    $_ -replace '\$env:', '' -replace '^RS_', ''
  }
  & $NSSM set RightLink AppEnvironmentExtra $RS_VARS
  & $NSSM start RightLink
} else {
  LogWrite "Unable to start RightLink service, no enrollment tokens found at $RS_ID_FILE"
}

# We wait forever here so that NSSM can treat this as a "service"
LogWrite "Userdata fetcher sitting in loop to simulate a service"
Get-Job | Remove-Job -Force
While($true) { Start-Sleep -Seconds 3600 }
