# Configure Networks accroding to the provice metadata
#

Param(
    [switch]$Help,
    [switch]$Debug
)

# Trap is more like "catch" in that is only catches exceptions from throw and not exits
trap {
    LogWrite "Caught Exception $_"
    StartUserDataService
    exit 1
}

function GetHelp {
    Write-Host "Configure networking for RightScale Cloud Appliance for vScale (RCA-V)."
    Write-Host "Parameters:"
    Write-Host "  -Debug                Verbose output"
    Write-Host "  -Help                 Display help"
    Write-Host ""
}

$Debug = $true

if ($Debug) {
    $DebugPreference = "Continue"
}

if ($Help) {
    GetHelp
    exit 0
}

function LogWrite([string]$logString) {
    $currentTime = Get-Date -Format "yyyy-MM-dd hh:mm:ss"
    Add-Content $NETWORK_LOG_FILE "[$currentTime] INFO $logString`n"
}

function LogError([string]$logString) {
    $currentTime = Get-Date -Format "yyyy-MM-dd hh:mm:ss"
    Add-Content $NETWORK_LOG_FILE "[$currentTime] ERR  $logString`n"
    throw $logString
}

function LogDebug([string]$logString) {
    if ($Debug) {
        LogWrite("DEBUG: $logString")
    }
}

function StartUserDataService() {
    $RIGHTLINK_DIR = "$env:ProgramFiles\RightScale\RightLink"
    $NSSM          = "$RIGHTLINK_DIR\nssm.exe"
    # RightScaleUserData service should always start RightLink when done
    LogWrite "Starting RightScaleUserdata service"
    & $NSSM start RightScaleUserData
}


function Query-Metadata ($type) {
    $metadata = (& $RPCTOOL_BIN "info-get guestinfo.$type" 2>&1)
    if (!$?) {
        if ($metadata -contains "No value found") {
            return ""
        }
        LogDebug "Fail to fetch $type"
        return $null
    }
    return $metadata
}

function Get-Metadata {
    $RETRIES = 20
    $metadata = $null
    LogWrite "Fetching metadata"
    do {
        $metadata = Query-Metadata "metadata"
        if (($metadata -eq $null) -or ($metadata -match '=\s*discovering(&|$)')) {
            $RETRIES--
            Start-Sleep -Seconds 30
        } else {
            LogWrite 'Done'
            return $metadata
        }
    } while ($RETRIES -gt 0)
    LogWrite 'Fail'
    return $false
}

function Get-Network-Settings($metadata) {
    $settings = @{ "networks" = @{}; "routes" = @{} }
    LogWrite 'Got following metadata:'
    foreach ($line in ($metadata | foreach {$_ -split "&|`n"} )) {
        $line = $line -replace "(.*)='(.*)'", "`$1=`$2"
        $name, $value = $line -split "="
        $name = $name.ToUpper()
        if ($name -notmatch '^RS_') {
            # Skip non RS keys
            continue
        }
        LogWrite "  $name = $value"

        if ($name.Contains("RS_IP") ) {
            $name = $name.Replace("RS_IP", "")
            $index, $key = $name.Split('_', 2)
            $network = $settings["networks"][$index]
            if ($network -eq $null ) {
                $network = @{}
            }
            $network[$key] = $value
            $settings["networks"][$index] = $network

        } elseif ($name.Contains("RS_ROUTE")) {
            $index = $name.Replace("RS_ROUTE", "")
            $route = $settings["routes"][$index]
            if($route -eq $null) {
                $route = @{}
            }
            $route["NAT_SERVER_IP"], $route["CIDR"] = $value.Split(',|;', 2)
            $settings["routes"][$index] = $route
        }
    }
    return $settings
}

function Get-Network-Devices {
    [array] $devices = $null
    LogDebug "=== Get Network Devices"
    LogDebug "Network Devices:"
    Get-WmiObject win32_networkadapter | % {
        if(($_.MACAddress) -ne $null -and ($_.NetConnectionID) -ne $null) {
            LogDebug "  $($_.MACAddress) $($_.NetConnectionID)"
            $devices += $_
        }
    }
    return $devices
}

function Get-Default-Network-Devices-Names {
    [String[]] $names = (1..10)
    $names[0] = "Ethernet"
    for($i = 2; $i -le $names.Count; $i++) {
        $names[$i-1] = "Ethernet $i"
    }
    return $names
}

function Get-Static-Indexes($networkSettings) {
    [array] $indexes = $null
    $networkSettings.Keys | % {
        $settings = $networkSettings.Item($_)
        if  ($settings["ADDR"] -ne $null) {
          $indexes = $indexes + $_
        }
    }
    return $indexes
}

function Get-Mac-Indexes($networkSettings) {
    [array] $indexes = $null
    $renameRequire = $false
    $networkSettings.Keys | % {
        $settings = $networkSettings.Item($_)
        if  ($settings["MAC"] -ne $null) {
            $indexes = $indexes + $_;
            $renameRequire = ($renameRequire -or ($settings["NAME"] -ne $null));
        }
    }
    if ($renameRequire) {
        return $indexes
    }
    return $null
}

function Get-Gateway-Indexes($networkSettings) {
    [array] $indexes = $null
    $networkSettings.Keys | % {
        $settings = $networkSettings.Item($_)
        if  ($settings["GATEWAY"] -ne $null) {
          $indexes = $indexes + $_
        }
    }
    return $indexes
}

function Get-Device-Name-From-Mac([string] $mac) {
    [string] $name = Get-Network-Devices | % {
        if ($_.MACAddress.ToUpper() -eq $mac.ToUpper()) {
            return $_.NetConnectionID
        }
    }
    if ( $name -eq $null) {
        LogError "Could not find Device for $mac"
    }
    return $name
}

function Get-Nameservers-Array([string] $nameservers) {
    $nameservers.Replace(' ', ',').Split(',',[System.StringSplitOptions]::RemoveEmptyEntries)
}

function Get-Device-IP($device) {
    $cmd = "netsh interface ipv4 show addresses `"$device`""
    $result = Invoke-Expression $cmd | Select-String -Pattern 'IP Address:\s+([0-9\.]+)'
    if (($result = Invoke-Expression $cmd | Select-String -Pattern 'IP Address:\s+([0-9\.]+)') -ne $null) {
        return $result.Matches[0].Groups[1].Value
    }
    return $null
}

function Wait-Configuration-Appliance($device, $ipaddr) {
    LogWrite "Waiting configuration appliance for `"$device`" with IP $ipaddr "
    do {
        Sleep(2)
    } while ($ipaddr -ne (Get-Device-IP $device))
    LogWrite "Done"
}

function Get-Configured-Nameservers($device) {
    $cmd = "netsh interface ip show dns `"$device`""
    LogDebug $cmd
    return (
        (Invoke-Expression $cmd) -split "`n" | % {
            if ($_ -notmatch 'DHCP') {
                $_
            }
        }
    ) -join ''
}

function Is-All-Nameservers-Configured($device, $nameservers) {
    $configured = Get-Configured-Nameservers($device)
    $isAllConfigured = $true
    foreach($nameserver in $nameservers) {
        $isAllConfigured = ($isAllConfigured -and $configured -like "*$nameserver*")
        if (!$isAllConfigured) {
            break
        }
    }
    return $isAllConfigured
}

function Add-Static-Ip($settings) {
    [string] $ipaddr = $settings["ADDR"]
    [string] $netmask = $settings["NETMASK"]
    [string] $gateway = $settings["GATEWAY"]
    [string] $device = Get-Device-Name-From-Mac $settings["MAC"]
    [array] $nameservers = Get-Nameservers-Array $settings["NAMESERVERS"]

    if ($ipaddr -eq "") {
        return $false
    }

    LogWrite "Configure Static address $ipaddr for `"$device`""
    LogWrite "Netmask: $netmask, Gateway: $gateway"

    if ($gateway -eq "") {
        $gateway = "none"
    } else {
        $gateway = "$gateway gwmetric=1"
    }

    $cmd = "netsh interface ip set address name=`"$device`" source=static addr=$ipaddr mask=$netmask gateway=$gateway"
    LogWrite $cmd
    (Invoke-Expression $cmd > $null)

    Wait-Configuration-Appliance $device $ipaddr

    if (($nameservers -eq $null) -or (Is-All-Nameservers-Configured $device $nameservers)) {
        LogWrite "Nameservers is missing or already configured for `"$device`""
        return
    }

    LogWrite "Configure Nameservers: $($nameservers -join ", ") for `"$device`""

    for($index = 1; $index -le $nameservers.Count; $index++) {
        $nameserver = $nameservers[$index-1]
        if ($index -eq 1) {
            $cmd = "netsh interface ipv4 set dnsserver name=`"$device`" source=static addr=$nameserver register=primary validate=no"
        } else {
            $cmd = "netsh interface ipv4 add dnsserver name=`"$device`" addr=$nameserver index=$index validate=no"
        }
        LogWrite $cmd
        (Invoke-Expression $cmd > $null)
    }
}

function Convert-CIDR($CIDRIPAddress) {
    $ipaddr, $cidr = $CIDRIPAddress.Split('/',2)

    if($cidr -le 32){
        [Int[]]$array = (1..32)
        for($i=0;$i -lt $array.length;$i++){
            if($array[$i] -gt $cidr){
                $array[$i]="0"
            } else {
                $array[$i]="1"
            }
        }
        $binary = $array -join ""
    }

    $i = 0
    do {
        $netmask += "." + [string]$([convert]::toInt32($binary.substring($i,8),2));
        $i+=8
    } while ($i -le 24)
    return $ipaddr, $netmask.Substring(1)
}

function Add-Network-Route($network, $nat_server_ip) {
    LogWrite "Configuring route for $network to $nat_server_ip"

    $ipaddr, $netmask = Convert-CIDR $network

    $cmd = "route print"
    LogDebug $cmd
    $result = (Invoke-Expression $cmd | Select-String -Pattern "$ipaddr.*$netmask.*$nat_server_ip" )

    if ($result -ne $null) {
      LogWrite "Route $ipaddr $netmask $nat_server_ip already exists"
      return
    }

    $cmd = "route -p ADD $ipaddr MASK $netmask $nat_server_ip"
    LogWrite $cmd
    Invoke-Expression $cmd > $null
}

function Get-Safe-Name($name) {
    return [regex]::Replace($name, '[<>:/\\|\?*"]', '.')
}

function Do-Network-Rename($oldName , $newName) {
    $cmd = "netsh interface set interface `"$oldName`"  newname=`"$newName`""
    LogWrite $cmd
    (Invoke-Expression $cmd > $null)
}

function Get-Default-Gateway-Settings ($networkSettings) {
    $indexes = Get-Gateway-Indexes $networkSettings
    if ($indexes -eq $null) {
        return $null
    }
    # Because Strings are not sorted as Ints, and we need to find smallest Int.
    # We are looping through all Strings and covert them to Int and compare
    # By default first String in array (converted to Int) is smallest
    $smallest = -1
    $isFirst = $true
    foreach($index in $indexes) {
        $temp = [convert]::ToInt16($index);
        if ($isFirst -or ($temp -lt $smallest)) {
            $smallest = $temp
            $isFirst = $false
        }
    }
    return $networkSettings[[string]$smallest]
}

function Configure-Default-Gateway($settings) {
    LogWrite "Configure with following gateway $($settings["GATEWAY"])"
    $cmd = "netsh interface ipv4 set interface `"$($settings["NAME"])`" METRIC=0"
    LogDebug $cmd
    Invoke-Expression $cmd > $null
}

###############################################
# Begin Main
###############################################
$RIGHTNETWORK_DIR       = "$env:ProgramFiles\RightScale\RightNetwork"
$LOG_DIR                = "$RIGHTNETWORK_DIR\Logs"
$NETWORK_LOG_FILE       = "$LOG_DIR\rightscale_network.log"
$RPCTOOL_BIN            = "C:\Program Files\VMware\VMware Tools\rpctool.exe"

if (!(Test-Path -Path $RPCTOOL_BIN)) {
  LogError "Could not find '$RPCTOOL_BIN'"
}

New-Item -Path $LOG_DIR -Type Directory -Force | Out-Null

# If log file doesn't exist then create it and the directory
if (!(Test-Path -Path $NETWORK_LOG_FILE)) {
  New-Item -Path $NETWORK_LOG_FILE -Type File -Force | Out-Null
}

LogWrite "=== Fetching Cloud Meta Data"
if (($metadata = Get-Metadata) -eq $false) {
  LogError "No valid metadata sources found"
}
LogDebug $metadata

LogWrite "=== Setup Default Network Devices Names"
$OS_NET_DEVICES = Get-Default-Network-Devices-Names
$OS_NET_DEVICES_INDEX = 0

$settings = Get-Network-Settings $metadata

if (($indexes = Get-Static-Indexes  $settings["networks"]) -ne $null) {
    LogWrite "=== Configure Static IPs"
    foreach($index in $indexes) {
        Add-Static-Ip $settings["networks"][$index]
    }
}

if ($settings["routes"].Count -gt 0) {
    LogWrite "=== Configure Routes"
    foreach($route in $settings["routes"].GetEnumerator()) {
        Add-Network-Route $route.Value["CIDR"] $route.Value["NAT_SERVER_IP"]
    }
}

# VMware may give us the network adaptors out of order. The primary network may come in as the the second
# device and vice versa. Rename devices to nice stable names based on the network name as a nicety.
# We do a two step process - rename to the MAC and rename to final name.
if (($indexes = Get-Mac-Indexes $settings["networks"]) -ne $null) {
    LogWrite "=== Rename Adapters"

    # The sort-object is critical - It ensures the longest names come first. The renaming command
    # (netsh interface set interface) does a partial match on the name, so if the have "Ethernet" and "Ethernet 2" and
    # try to rename Ethernet first, it will rename either "Ethernet" or "Ethernet 2" at random since both names 
    # match the partial match.
    Get-Network-Devices |  Sort-Object -Property 'NetConnectionID' -descending | % {
        $tempName = $(Get-Safe-Name $_.MACAddress).toUpper()
        $deviceName = $_.NetConnectionID
        Do-Network-Rename $deviceName $tempName
    }
    foreach($index in $indexes) {
        $networkSettings = $settings["networks"][$index]
        $tempName = $(Get-Safe-Name $networkSettings["MAC"]).toUpper()
        if ($networkSettings["NAME"] -ne $null) {
            $deviceName = Get-Safe-Name $networkSettings["NAME"]
        } else {
            $deviceName = $OS_NET_DEVICES[$OS_NET_DEVICES_INDEX++]
        }
        Do-Network-Rename $tempName $deviceName
    }
}

if(($networkSettings = Get-Default-Gateway-Settings $settings["networks"]) -ne $null) {
    LogWrite "=== Configure Default Gateway"
    Configure-Default-Gateway $networkSettings
}

StartUserDataService

